/*
    Ishtar - A network transparent read/write interface with
    probing and enumeration support.
    Copyright (C) 2003-2005 Stephane Magnenat <nct@ysagoon.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <ishtar/services.h>
#include <stdio.h>

using namespace Ishtar;

class TextClient : public NetworkClient
{
	typedef std::map<std::string, UInt32> VarMap;
	VarMap varMap;

	typedef std::map<UInt32, ServiceDescription *> ServiceDescriptionMap;
	ServiceDescriptionMap serviceDescriptionMap;

public:

	TextClient(const std::string &host, unsigned port)
	:NetworkClient(host, port)
	{
		// Print some infos
		std::cout << "Ishtar text client, using Ishtar Protocol Version " << PROTOCOL_VERSION << std::endl;
		// Send hello
		HelloCall hello(HelloCall::CLIENT);
		hello.serialize(socket);
		socket->flush();

		// Ask for service list
		GetServicesCall getServiceCall;
		getServiceCall.serialize(socket);
		socket->flush();
	}

	// Return true if we wait an answer
	bool handelInput(void)
	{
		char buf[1024];
		if (fgets(buf,1024,stdin))
		{
			Call *call;
			char *cmd = NULL;
			char *params = NULL;

			cmd=strtok(buf," \t\r\n");
			params=strtok(NULL,"\r\n");
			
			if (cmd == NULL)
				return false;

			if (strcmp(cmd, "l")==0)
			{
				GetServicesCall	getServiceCall;
				getServiceCall.serialize(socket);
				socket->flush();
				return true;
			}
			else if (strcmp(cmd, "g")==0)
			{
				// first we get values from string
				char *nt=strtok(params," \t\r\n");
				char *val[64];
				unsigned count=0;
				while (nt && count<64)
				{
					val[count++]=nt;
					nt=strtok(NULL," \t\r\n");
				}

				// we now have in val[0..count-1] our values
				if (!count)
					return false;

				GetValuesCall getValuesCall;
				for (unsigned i=0; i<count; i++)
				{
					VarMap::const_iterator e=varMap.find(val[i]);
					if (e!=varMap.end())
					{
						getValuesCall.services.push_back(e->second);
					}
					else
					{
						std::cerr << "No service named " << val[i] << std::endl;
					}
				}

				if (getValuesCall.services.size())
				{
					getValuesCall.serialize(socket);
					socket->flush();
					return true;
				}
			}
			else if (strcmp(cmd, "s")==0)
			{
				// first we get values from string
				char *nt=strtok(params," \t\r\n");
				char *val[64];
				unsigned count=0;
				while (nt && count<64)
				{
					val[count++]=nt;
					nt=strtok(NULL," \t\r\n");
				}

				// we now have in val[0..count-1] our values
				if (!count)
					return false;

				VarMap::const_iterator e=varMap.find(val[0]);
				if (e != varMap.end())
				{
					UInt32 sid=e->second;
					ServiceDescription *sd=serviceDescriptionMap[sid];
					if (!(sd->flags & ServiceDescription::READ_ONLY))
					{
						if ((sd->type == ValueVector::UCHAR) && (count==2))
						{
							// We have string
							SetValuesCall setValuesCall;
							setValuesCall.values[sid] = new CharValueVector(val[1]);

							setValuesCall.serialize(socket);
							socket->flush();

							setValuesCall.deleteValues();
						}
						else if ((sd->length>0) && (sd->length != count-1))
						{
							std::cerr << "Wrong number of arguments" << std::endl;
						}
						else
						{
							SetValuesCall setValuesCall;

							setValuesCall.values[sid] = ValueVector::newOfType(sd->type, count-1);

							for (unsigned i=1; i<count; i++)
							{
								setValuesCall.values[sid]->fromString(i-1, val[i]);
							}

							setValuesCall.serialize(socket);
							socket->flush();

							setValuesCall.deleteValues();
						}
					}
					else
					{
						std::cerr << "Service " << val[0] << " is read-only" << std::endl;
					}
				}
				else
				{
					std::cerr << "No service named " << val[0] << std::endl;
				}
			}
			else if(strcmp(cmd, "q")==0)
			{
				exit(0);
			}
			else
			{
				std::cerr << "Unknown command" << std::endl;
			}
		}
		else
		{
			// TODO : quit
		}
		return false;
	}

	void dumpServiceDescription(UInt32 id, const ServiceDescription *sd)
	{
		std::string name = sd->name;
		std::cout << id << " : " << name;
		if (sd->flags & ServiceDescription::READ_ONLY)
			std::cout << " (ro)";
		if (sd->length > 0)
		{
			std::cout << "\n";
			if ((sd->flags & ServiceDescription::NAMED_VALUES) || (sd->flags & ServiceDescription::CONSTRAINT_VALUES))
			{
				for (unsigned i=0; i<sd->length; i++)
				{
					std::cout << "\t";
					if (sd->flags & ServiceDescription::NAMED_VALUES)
					{
						std::string subname = sd->subnames[i];
						std::cout << subname;
					}
					if (sd->flags & ServiceDescription::CONSTRAINT_VALUES)
					{
						std::cout << " / min = " << sd->mins->toString(i);
						std::cout << " / max = " << sd->maxs->toString(i);
					}
					std::cout << "\n";
				}
			}
		}
		else
			std::cout << " (dynamic length)\n";
		std::cout << std::endl;
	}

	void incomingData(Socket *socket)
	{
		std::cout << "incomingData" << std::endl;
		Answer *answer = Answer::deserialize(socket);

		ServiceAnswer *serviceAnswer = dynamic_cast<ServiceAnswer *>(answer);
		if (serviceAnswer)
		{
			std::cout << "ServiceAnswer : \n";
			for (ServiceAnswer::DescriptionMap::const_iterator it = serviceAnswer->descriptions.begin(); it != serviceAnswer->descriptions.end(); ++it)
			{
				dumpServiceDescription(it->first, it->second);
				serviceDescriptionMap[it->first] = it->second;
				varMap[it->second->name] = it->first;
			}
		}

		ValuesAnswer *valuesAnswer = dynamic_cast<ValuesAnswer *>(answer);
		if (valuesAnswer)
		{
			std::cout << "ValuesAnswer :\n";
			for (ValuesAnswer::ValueVectorMap::const_iterator it = valuesAnswer->values.begin(); it != valuesAnswer->values.end(); ++it)
			{
				const ServiceDescription *sd = serviceDescriptionMap[it->first];
				std::string sdName = sd->name;
				std::cout << sdName;
				if (sd->length == 0)
					std::cout << "(dynamic length)";
				else if (sd->length != it->second->size())
					std::cout << "(warning: missmatching length)";
				std::cout << " :\n";
				for (unsigned i = 0; i<it->second->size(); i++)
				{
					std::cout << it->second->toString(i) << " ";
				}
				std::cout << "\n";
			}
			std::cout << std::endl;

			valuesAnswer->deleteValues();
		}
		
		ErrorAnswer *errorAnswer = dynamic_cast<ErrorAnswer *>(answer);
		if (errorAnswer)
		{
			std::cout << "Error " << errorAnswer->errorType << "\n" << std::endl;
		}

		delete answer;

		while (!handelInput());
	}

	void connectionClosed(Socket *socket)
	{
		std::cout << "Connection closed to " << socket->getRemoteName() << std::endl;
	}
};

int main(int argc, char *argv[])
{
	if (argc!=3)
	{
		std::cerr << "Usage : " << argv[0] << " hostname port" << std::endl;
		return 1;
	}
	TextClient nc(argv[1], atoi(argv[2]));

	nc.run();

	return 0;
}
